# 10/17 CSC 401 Examples


# Load the dataset
data(iris)
set.seed(123)

# Convert species to a factor (for classification)
iris$Species <- as.factor(iris$Species)

# Scale the predictors
iris[, 1:4] <- scale(iris[, 1:4])

# Split the data into training and testing sets (70% train, 30% test)
train_indices <- sample(1:nrow(iris), 0.7 * nrow(iris))
train_data <- iris[train_indices, ]
test_data <- iris[-train_indices, ]

# Manhattan distance function
manhattan_distance <- function(x1, x2) {
  sum(abs(x1 - x2))
}

# Minkowski distance function with p = 3
minkowski_distance <- function(x1, x2, p = 3) {
  sum(abs(x1 - x2) ^ p)^(1/p)
}

# Chebyshev distance function
chebyshev_distance <- function(x1, x2) {
  max(abs(x1 - x2))
}

# General KNN function
knn_general <- function(train_data, test_point, k = 3, distance_func) {
  distances <- apply(train_data[, -ncol(train_data)], 1, function(row) distance_func(row, test_point))
  sorted_indices <- order(distances)
  nearest_neighbors <- train_data[sorted_indices[1:k], ]
  
  # Return the majority vote for classification
  prediction <- names(sort(table(nearest_neighbors[, ncol(nearest_neighbors)]), decreasing = TRUE))[1]
  
  return(prediction)
}

# Function to calculate evaluation metrics
evaluate_knn <- function(train_data, test_data, k, distance_func) {
  predictions <- sapply(1:nrow(test_data), function(i) {
    knn_general(train_data, test_data[i, -ncol(test_data)], k, distance_func)
  })
  
  true_labels <- test_data$Species
  
  # Calculate confusion matrix
  confusion_matrix <- table(Predicted = predictions, Actual = true_labels)
  
  # Calculate accuracy
  accuracy <- sum(diag(confusion_matrix)) / sum(confusion_matrix)
  
  # Calculate F1 score (macro-averaged)
  f1_score <- mean(sapply(levels(true_labels), function(class) {
    precision <- confusion_matrix[class, class] / sum(confusion_matrix[class, ])
    recall <- confusion_matrix[class, class] / sum(confusion_matrix[, class])
    f1 <- ifelse(precision + recall > 0, 2 * precision * recall / (precision + recall), 0)
    return(f1)
  }))
  
  list(predictions = predictions, accuracy = accuracy, f1_score = f1_score, confusion_matrix = confusion_matrix)
}

# Evaluate with Manhattan Distance
results_manhattan <- evaluate_knn(train_data, test_data, k = 5, distance_func = manhattan_distance)
results_manhattan$accuracy
results_manhattan$f1_score
results_manhattan$confusion_matrix

# Evaluate with Minkowski Distance (p = 3)
results_minkowski <- evaluate_knn(train_data, test_data, k = 5, distance_func = function(x1, x2) minkowski_distance(x1, x2, p = 3))
results_minkowski$accuracy
results_minkowski$f1_score
results_minkowski$confusion_matrix

# Evaluate with Chebyshev Distance
results_chebyshev <- evaluate_knn(train_data, test_data, k = 5, distance_func = chebyshev_distance)
results_chebyshev$accuracy
results_chebyshev$f1_score
results_chebyshev$confusion_matrix

# Prepare data for plotting
plot_data <- data.frame(
  Metric = rep(c("Accuracy", "F1 Score"), each = 3),
  Value = c(results_manhattan$accuracy, results_minkowski$accuracy, results_chebyshev$accuracy,
            results_manhattan$f1_score, results_minkowski$f1_score, results_chebyshev$f1_score),
  Distance = factor(rep(c("Manhattan", "Minkowski", "Chebyshev"), 2), levels = c("Manhattan", "Minkowski", "Chebyshev"))
)

# Plot the results
library(ggplot2)

ggplot(plot_data, aes(x = Distance, y = Value, fill = Metric)) +
  geom_bar(stat = "identity", position = "dodge") +
  theme_minimal() +
  labs(title = "Comparison of KNN with Different Distance Metrics", y = "Score", x = "Distance Metric") +
  scale_fill_manual(values = c("Accuracy" = "blue", "F1 Score" = "red"))


#mtcars example
# Load the mtcars dataset
data(mtcars)

# Convert 'am' to a factor for classification
mtcars$am <- as.factor(mtcars$am)

# Select the numeric predictor variables and normalize them
mtcars[, c("mpg", "hp")] <- scale(mtcars[, c("mpg", "hp")])

# Split the data into training and testing sets (70% train, 30% test)
set.seed(123)
train_indices <- sample(1:nrow(mtcars), 0.7 * nrow(mtcars))
train_data <- mtcars[train_indices, c("mpg", "hp", "am")]
test_data <- mtcars[-train_indices, c("mpg", "hp", "am")]

euclidean_distance <- function(x1, x2) {
  sqrt(sum((as.numeric(x1) - as.numeric(x2)) ^ 2))
}

cosine_similarity <- function(x1, x2) {
  sum(as.numeric(x1) * as.numeric(x2)) / (sqrt(sum(as.numeric(x1)^2)) * sqrt(sum(as.numeric(x2)^2)))
}

knn_classifier <- function(train_data, test_point, k = 3, distance_func) {
  distances <- apply(train_data[, -ncol(train_data)], 1, function(row) distance_func(row, test_point))
  
  # Determine whether to sort distances ascending or descending
  if (identical(distance_func, cosine_similarity)) {
    nearest_neighbors <- order(distances, decreasing = TRUE)
  } else {
    nearest_neighbors <- order(distances)
  }
  
  nearest_neighbors <- train_data[nearest_neighbors[1:k], "am"]
  
  # Majority voting
  prediction <- as.character(sort(table(nearest_neighbors), decreasing = TRUE)[1])
  return(prediction)
}

predict_knn <- function(train_data, test_data, k = 3, distance_func) {
  predictions <- sapply(1:nrow(test_data), function(i) {
    knn_classifier(train_data, test_data[i, -ncol(test_data)], k, distance_func)
  })
  return(as.factor(predictions))
}

# Predictions using Euclidean distance
predictions_euclidean <- predict_knn(train_data, test_data, k = 3, distance_func = euclidean_distance)

# Predictions using Cosine similarity
predictions_cosine <- predict_knn(train_data, test_data, k = 3, distance_func = cosine_similarity)

library(caret)

evaluate_model <- function(predictions, true_labels) {
  confusion <- confusionMatrix(predictions, true_labels)
  accuracy <- confusion$overall["Accuracy"]
  f1 <- confusion$byClass["F1"]
  
  return(list(Confusion_Matrix = confusion$table, Accuracy = accuracy, F1_Score = f1))
}

library(caret)

evaluate_model <- function(predictions, true_labels) {
  # Ensure that predictions and true labels have the same levels
  levels(predictions) <- levels(true_labels)
  
  # Calculate confusion matrix, accuracy, and F1 score
  confusion <- confusionMatrix(predictions, true_labels)
  accuracy <- confusion$overall["Accuracy"]
  f1 <- confusion$byClass["F1"]
  
  return(list(Confusion_Matrix = confusion$table, Accuracy = accuracy, F1_Score = f1))
}

# Euclidean results
euclidean_results <- evaluate_model(predictions_euclidean, test_data$am)
print("Euclidean Results")
print(euclidean_results)

# Cosine results
cosine_results <- evaluate_model(predictions_cosine, test_data$am)
print("Cosine Results")
print(cosine_results)

library(pROC)

plot_roc_auc <- function(predictions, true_labels, title) {
  roc_obj <- roc(as.numeric(true_labels) - 1, as.numeric(predictions) - 1)
  auc_value <- auc(roc_obj)
  
  plot(roc_obj, main = paste("ROC Curve:", title))
  return(auc_value)
}

# ROC and AUC for Euclidean
auc_euclidean <- plot_roc_auc(predictions_euclidean, test_data$am, "Euclidean Distance")

# ROC and AUC for Cosine
auc_cosine <- plot_roc_auc(predictions_cosine, test_data$am, "Cosine Similarity")

# Prepare data
set.seed(123)
mtcars$am <- as.factor(mtcars$am)  # Convert 'am' to a factor (for classification)

# Use 'hp', 'wt', and 'qsec' as predictors and 'am' as the target
data <- mtcars[, c("hp", "wt", "qsec", "am")]

# Scale the predictors
data[, 1:3] <- scale(data[, 1:3])

# Euclidean distance function
euclidean_distance <- function(x1, x2) {
  sqrt(sum((x1 - x2) ^ 2))
}

# KNN function using Euclidean distance
knn_euclidean <- function(train_data, test_point, k = 3) {
  distances <- apply(train_data[, -ncol(train_data)], 1, function(row) euclidean_distance(row, test_point))
  sorted_indices <- order(distances)
  nearest_neighbors <- train_data[sorted_indices[1:k], ]
  
  # Return the majority vote for classification
  prediction <- names(sort(table(nearest_neighbors[, ncol(nearest_neighbors)]), decreasing = TRUE))[1]
  
  return(as.numeric(prediction))
}

# Split the data into train and test sets
train_data <- data[-1, ]  # Use all rows except the first as training data
test_point <- data[1, -4]  # Use the first row (without the target column) as the test point

# Predict using KNN with Euclidean distance
predicted_value <- knn_euclidean(train_data, test_point, k = 3)
predicted_value

# Cosine similarity function
cosine_similarity <- function(x1, x2) {
  sum(x1 * x2) / (sqrt(sum(x1 ^ 2)) * sqrt(sum(x2 ^ 2)))
}

# Split the data into train and test sets
train_data <- data[-1, ]  # Use all rows except the first as training data
test_point <- data[1, -4]  # Use the first row (without the target column) as the test point

# Predict using KNN with Euclidean distance
predicted_value <- knn_euclidean(train_data, test_point, k = 3)
predicted_value

# KNN function using Cosine Similarity
knn_cosine <- function(train_data, test_point, k = 3) {
  similarities <- apply(train_data[, -ncol(train_data)], 1, function(row) cosine_similarity(row, test_point))
  sorted_indices <- order(-similarities)  # Sort in decreasing order because higher cosine similarity is better
  nearest_neighbors <- train_data[sorted_indices[1:k], ]
  
  # Return the majority vote for classification
  prediction <- names(sort(table(nearest_neighbors[, ncol(nearest_neighbors)]), decreasing = TRUE))[1]
  
  return(as.numeric(prediction))
}

# Predict using KNN with Cosine similarity
predicted_value_cosine <- knn_cosine(train_data, test_point, k = 3)
predicted_value_cosine

