# 10/31 CSC 401 Examples

# Install and load necessary libraries
if (!require(caret)) install.packages("caret", dependencies=TRUE)
library(rpart)
library(caret)
library(dplyr)
library(ggplot2)

# Function to build a single decision tree on bootstrapped data
build_tree <- function(data, formula) {
  # Bootstrap sample
  bootstrapped_data <- data[sample(1:nrow(data), nrow(data), replace = TRUE), ]
  # Train a decision tree using rpart
  tree_model <- rpart(formula, data = bootstrapped_data, method = "class", control = rpart.control(minsplit = 2, cp = 0))
  return(tree_model)
}

# Function to predict using the Random Forest
predict_random_forest <- function(trees, test_data) {
  predictions <- sapply(trees, function(tree) predict(tree, newdata = test_data, type = "class"))
  # Aggregate predictions by majority voting
  final_predictions <- apply(predictions, 1, function(row) names(sort(table(row), decreasing = TRUE)[1]))
  return(final_predictions)
}

# Random Forest function
random_forest <- function(train_data, test_data, formula, n_trees = 100) {
  trees <- list()
  for (i in 1:n_trees) {
    tree <- build_tree(train_data, formula)
    trees[[i]] <- tree
  }
  
  # Make predictions on the test set
  predictions <- predict_random_forest(trees, test_data)
  
  return(list(predictions = predictions))
}

# Split the iris dataset
set.seed(123)
train_index <- createDataPartition(iris$Species, p = 0.7, list = FALSE)
train_data <- iris[train_index, ]
test_data <- iris[-train_index, ]

# Run the Random Forest
set.seed(123)
rf_results <- random_forest(train_data, test_data, Species ~ ., n_trees = 100)

# Predictions
rf_predictions <- rf_results$predictions

# Confusion Matrix and Metrics
conf_matrix <- confusionMatrix(factor(rf_predictions, levels = levels(test_data$Species)), test_data$Species)
print(conf_matrix)


# Load the necessary package for visualization
if (!require(pheatmap)) install.packages("pheatmap", dependencies=TRUE)
library(pheatmap)

# Create a heatmap of the confusion matrix
conf_matrix_df <- as.data.frame.matrix(conf_matrix$table)
pheatmap(conf_matrix_df, 
         cluster_rows = FALSE, 
         cluster_cols = FALSE,
         display_numbers = TRUE,
         number_format = "%.0f",
         color = colorRampPalette(c("white", "blue"))(100),
         main = "Confusion Matrix Heatmap")

#adaboost
# Load necessary libraries
if (!require(rpart)) install.packages("rpart", dependencies=TRUE)
library(rpart)

# AdaBoost Algorithm
adaboost <- function(X, y, M = 10) {
  n <- nrow(X)
  weights <- rep(1/n, n)  # Initialize weights
  classifiers <- list()  # To store classifiers
  classifier_weights <- numeric(M)  # To store classifier weights
  
  for (m in 1:M) {
    # Train weak learner
    model <- rpart(as.factor(y) ~ ., data = X, weights = weights, method = "class", control = rpart.control(maxdepth = 1))
    
    # Make predictions
    predictions <- predict(model, X, type = "class")
    
    # Calculate error and classifier weight
    incorrect <- as.numeric(predictions != y)
    error <- sum(weights * incorrect) / sum(weights)
    
    alpha <- 0.5 * log((1 - error) / max(error, 1e-10))
    classifier_weights[m] <- alpha
    
    # Update weights
    weights <- weights * exp(-alpha * (2 * as.numeric(predictions == y) - 1))
    weights <- weights / sum(weights)  # Normalize weights
    
    classifiers[[m]] <- model  # Store the weak learner
  }
  
  return(list(classifiers = classifiers, classifier_weights = classifier_weights))
}

predict_adaboost <- function(model, X) {
  M <- length(model$classifiers)
  n <- nrow(X)
  
  # Initialize prediction matrix
  pred_matrix <- matrix(0, nrow = n, ncol = M)
  
  for (m in 1:M) {
    pred_matrix[, m] <- as.numeric(predict(model$classifiers[[m]], X, type = "class"))
  }
  
  # Calculate weighted vote (This needs to be adjusted)
  final_pred <- apply(pred_matrix, 1, function(row) {
    weighted_sum <- sum(model$classifier_weights * (2 * row - 3))  # 2*row-3 converts 1/-1 to 1/-1 scale for the vote.
    return(ifelse(weighted_sum > 0, 1, -1))  # Convert back to class labels
  })
  
  return(final_pred)
}

# Convert iris data into binary classification problem
iris_binary <- iris[iris$Species != "setosa", ]  # Remove setosa to make it binary
iris_binary$Species <- ifelse(iris_binary$Species == "versicolor", 1, -1)  # 1 for versicolor, -1 for virginica

# Check the balance of the dataset
table(iris_binary$Species)

# Retrain the model
set.seed(123)
ada_model <- adaboost(iris_binary[, -5], iris_binary$Species, M = 10)

# Predict on the training data
predictions <- predict_adaboost(ada_model, iris_binary[, -5])

# Confusion Matrix
conf_matrix <- table(Predicted = predictions, Actual = iris_binary$Species)
print(conf_matrix)

# Accuracy
accuracy <- sum(diag(conf_matrix)) / sum(conf_matrix)
print(paste("Accuracy:", round(accuracy, 3)))

# Visualization
roc_obj <- roc(iris_binary$Species, predictions)
plot(roc_obj, col = "blue", main = "ROC Curve for AdaBoost on Iris Data")
auc(roc_obj)

# Calculate confusion matrix
conf_matrix <- table(Predicted = predictions, Actual = iris_binary$Species)

# Ensure confusion matrix is 2x2 (it should be)
print(conf_matrix)

# Plot using fourfoldplot
fourfoldplot(conf_matrix, color = c("#CC6666", "#99CC99"),
             conf.level = 0.95, margin = 1, main = "Fourfold Plot of Confusion Matrix")
