# 11/14 CSC 401 Examples

# Load the iris dataset
data(iris)

# Extract the feature matrix and the true labels
iris_data <- iris[, 1:4]  # Only the numeric data
true_labels <- iris[, 5]  # The species (true labels)

set.seed(123)  # Set seed for reproducibility

# Euclidean distance function
euclidean_distance <- function(x, y) {
  sqrt(sum((x - y) ^ 2))
}

# K-Means algorithm
k_means <- function(data, k, max_iter = 100) {
  # Randomly initialize the centroids by selecting k points from the dataset
  centroids <- data[sample(1:nrow(data), k), ]
  
  # Create a vector to store the cluster assignments
  clusters <- rep(0, nrow(data))
  
  # Iterative optimization
  for (iter in 1:max_iter) {
    # Step 1: Assign clusters
    for (i in 1:nrow(data)) {
      distances <- apply(centroids, 1, function(centroid) euclidean_distance(data[i, ], centroid))
      clusters[i] <- which.min(distances)
    }
    
    # Step 2: Update centroids
    new_centroids <- matrix(NA, nrow = k, ncol = ncol(data))
    for (j in 1:k) {
      if (sum(clusters == j) == 0) {
        # In case a cluster is empty, reinitialize that centroid randomly
        new_centroids[j, ] <- data[sample(1:nrow(data), 1), ]
      } else {
        new_centroids[j, ] <- colMeans(data[clusters == j, , drop = FALSE])
      }
    }
    
    # Check for convergence (if centroids don't change)
    if (all(new_centroids == centroids)) {
      break
    }
    
    centroids <- new_centroids
  }
  
  list(centroids = centroids, clusters = clusters)
}

# Run the K-Means algorithm
k = 3
kmeans_result <- k_means(iris_data, k)

# Extract the cluster assignments
predicted_clusters <- kmeans_result$clusters

# Convert predicted_clusters to factor with levels matching true_labels
predicted_clusters <- as.factor(predicted_clusters)

# Assign appropriate labels to the predicted clusters
# This step requires a bit of manual inspection to determine which cluster corresponds to which species.
# Assuming the clusters 1, 2, 3 correspond to species "setosa", "versicolor", "virginica"
levels(predicted_clusters) <- levels(true_labels)

# Now create the confusion matrix
conf_matrix <- confusionMatrix(predicted_clusters, true_labels)

# Print the confusion matrix
print(conf_matrix)

# Calculate overall accuracy
accuracy <- sum(diag(conf_matrix$table)) / sum(conf_matrix$table)
cat("Overall Accuracy: ", accuracy, "\n")

# Assign species names based on which cluster has the majority of each species
library(dplyr)

# Create a mapping of clusters to species based on the majority vote
mapping <- data.frame(Cluster = predicted_clusters, Species = true_labels) %>%
  group_by(Cluster) %>%
  summarize(Majority_Species = names(which.max(table(Species))))

# Apply the mapping to the predicted clusters
predicted_labels <- mapping$Majority_Species[match(predicted_clusters, mapping$Cluster)]

# Convert predicted labels to a factor with the same levels as true labels
predicted_labels <- factor(predicted_labels, levels = levels(true_labels))

# Confusion matrix with updated labels
conf_matrix <- confusionMatrix(predicted_labels, true_labels)

# Print the confusion matrix
print(conf_matrix)

# Calculate overall accuracy
accuracy <- sum(diag(conf_matrix$table)) / sum(conf_matrix$table)
cat("Overall Accuracy: ", accuracy, "\n")

# PCA for visualization
pca_result <- prcomp(iris_data, scale. = TRUE)
iris_pca <- data.frame(pca_result$x[, 1:2], Cluster = as.factor(predicted_clusters), Species = true_labels)

# Plot the clusters
library(ggplot2)
ggplot(iris_pca, aes(x = PC1, y = PC2, color = Cluster, shape = Species)) +
  geom_point(size = 3) +
  labs(title = "K-Means Clustering on Iris Dataset",
       x = "Principal Component 1", y = "Principal Component 2") +
  theme_minimal() +
  theme(legend.position = "bottom")

#multiple runs
# Load necessary library
library(dplyr)
library(caret)
library(ggplot2)

# Define the Euclidean distance function
euclidean_distance <- function(x1, x2) {
  return(sqrt(sum((x1 - x2)^2)))
}

# Define the K-Means function
kmeans_custom <- function(data, k, max_iters = 100) {
  # Randomly select k initial centroids
  set.seed(Sys.time())  # Change the seed based on current time for randomness
  centroids <- data[sample(1:nrow(data), k), ]
  
  for (i in 1:max_iters) {
    # Assign clusters based on closest centroid
    clusters <- apply(data, 1, function(row) {
      distances <- apply(centroids, 1, euclidean_distance, x2 = row)
      return(which.min(distances))
    })
    
    # Calculate new centroids
    new_centroids <- sapply(1:k, function(cluster) {
      return(colMeans(data[clusters == cluster, ]))
    })
    
    new_centroids <- t(new_centroids)
    
    # Check for convergence (if centroids do not change)
    if (all(centroids == new_centroids)) {
      break
    }
    
    centroids <- new_centroids
  }
  
  return(list(clusters = clusters, centroids = centroids))
}

# Function to calculate the Within-Cluster Sum of Squares (WCSS)
calculate_wcss <- function(data, clusters, centroids) {
  wcss <- 0
  for (i in 1:nrow(data)) {
    cluster <- clusters[i]
    centroid <- centroids[cluster, ]
    wcss <- wcss + sum((data[i, ] - centroid)^2)
  }
  return(wcss)
}

# Multiple runs K-Means
multiple_runs_kmeans <- function(data, k, num_runs = 10, max_iters = 100) {
  best_wcss <- Inf
  best_clusters <- NULL
  best_centroids <- NULL
  
  for (run in 1:num_runs) {
    set.seed(run)  # Set a seed for reproducibility
    kmeans_result <- kmeans_custom(data, k, max_iters)
    wcss <- calculate_wcss(data, kmeans_result$clusters, kmeans_result$centroids)
    
    if (wcss < best_wcss) {
      best_wcss <- wcss
      best_clusters <- kmeans_result$clusters
      best_centroids <- kmeans_result$centroids
    }
  }
  
  return(list(clusters = best_clusters, centroids = best_centroids, wcss = best_wcss))
}

# Prepare the iris dataset (excluding the species "setosa" and using only numeric variables)
iris_data <- iris[iris$Species != "setosa", ]
iris_data <- iris_data %>% select(-Species)
true_labels <- factor(iris[iris$Species != "setosa", ]$Species)

# Apply K-Means with multiple runs
k <- 2  # Setting k=2 for simplicity
num_runs <- 10
kmeans_result <- multiple_runs_kmeans(iris_data, k, num_runs)

# Manually adjust predicted cluster labels to match true labels
predicted_clusters <- factor(kmeans_result$clusters, levels = 1:k)
levels(predicted_clusters) <- levels(true_labels)

# Confusion matrix comparison
conf_matrix <- confusionMatrix(predicted_clusters, true_labels)
print(conf_matrix)

# Visualize the best clustering result
iris_plot_data <- iris[iris$Species != "setosa", ]
iris_plot_data$Cluster <- predicted_clusters

ggplot(iris_plot_data, aes(x = Petal.Length, y = Petal.Width, color = Cluster)) +
  geom_point(size = 3) +
  labs(title = "K-Means Clustering Result", x = "Petal Length", y = "Petal Width") +
  theme_minimal()

# Output the best WCSS
cat("Best WCSS from multiple runs: ", kmeans_result$wcss, "\n")
