# 11/21 CSC 401 Examples

# Load necessary libraries
library(tm)
library(e1071)
library(caret)
library(pROC)

# Load dataset (e.g., SMS spam dataset)
# Download and unzip the dataset
url <- "https://archive.ics.uci.edu/ml/machine-learning-databases/00228/smsspamcollection.zip"
download.file(url, destfile = "sms_spam.zip")
unzip("sms_spam.zip", files = "SMSSpamCollection")

# Load the dataset into R
sms_data <- read.table("SMSSpamCollection", sep = "\t", header = FALSE, stringsAsFactors = FALSE, col.names = c("label", "text"))

# Check the first few rows
head(sms_data)

# Convert labels to factors
sms_data$label <- factor(sms_data$label)

library(tm)
library(SnowballC)

# Create a text corpus
corpus <- VCorpus(VectorSource(sms_data$text))

# Convert the text to lowercase
corpus <- tm_map(corpus, content_transformer(tolower))

# Remove punctuation
corpus <- tm_map(corpus, removePunctuation)

# Remove stopwords
corpus <- tm_map(corpus, removeWords, stopwords("english"))

# Perform stemming
corpus <- tm_map(corpus, stemDocument)

# Convert the corpus to a document-term matrix
dtm <- DocumentTermMatrix(corpus)

# Inspect the document-term matrix
inspect(dtm[1:5, 1:10])

# Split the data into training (75%) and test (25%) sets
set.seed(123)
train_index <- sample(seq_len(nrow(sms_data)), size = 0.75 * nrow(sms_data))
sms_train <- sms_data[train_index, ]
sms_test <- sms_data[-train_index, ]

dtm_train <- dtm[train_index, ]
dtm_test <- dtm[-train_index, ]

library(e1071)

# Convert the document-term matrices to binary matrices (presence/absence of terms)
convert_counts <- function(x) {
  x <- ifelse(x > 0, 1, 0)
  return(factor(x, levels = c(0, 1), labels = c("No", "Yes")))
}

sms_train_matrix <- apply(dtm_train, MARGIN = 2, convert_counts)
sms_test_matrix <- apply(dtm_test, MARGIN = 2, convert_counts)

# Train a Naive Bayes model
sms_classifier <- naiveBayes(sms_train_matrix, sms_train$label)

# Make predictions
sms_predictions <- predict(sms_classifier, sms_test_matrix)

library(caret)

# Confusion matrix
conf_matrix <- confusionMatrix(sms_predictions, sms_test$label)

# Print the confusion matrix and related metrics
print(conf_matrix)

# Extract metrics
accuracy <- conf_matrix$overall["Accuracy"]
precision <- conf_matrix$byClass["Pos Pred Value"]
recall <- conf_matrix$byClass["Sensitivity"]
f1_score <- 2 * ((precision * recall) / (precision + recall))

cat("Accuracy:", accuracy, "\n")
cat("Precision:", precision, "\n")
cat("Recall:", recall, "\n")
cat("F1 Score:", f1_score, "\n")

library(wordcloud)

# Word cloud for spam messages
spam_corpus <- Corpus(VectorSource(sms_data$text[sms_data$label == "spam"]))
wordcloud(spam_corpus, max.words = 100, colors = "red")

# Word cloud for ham messages
ham_corpus <- Corpus(VectorSource(sms_data$text[sms_data$label == "ham"]))
wordcloud(ham_corpus, max.words = 100, colors = "blue")

library(tm)

# Sample text data (corpus)
texts <- c("I love machine learning", 
           "Machine learning is fun", 
           "I love learning about data science")

# Create a text corpus
corpus <- VCorpus(VectorSource(texts))

# Preprocess the text (convert to lowercase, remove punctuation, etc.)
corpus <- tm_map(corpus, content_transformer(tolower))
corpus <- tm_map(corpus, removePunctuation)
corpus <- tm_map(corpus, removeNumbers)
corpus <- tm_map(corpus, removeWords, stopwords("english"))

# Create the Document-Term Matrix (DTM)
dtm <- DocumentTermMatrix(corpus)

# Inspect the Document-Term Matrix
inspect(dtm)

# Create the DTM with TF-IDF weighting
dtm_tfidf <- DocumentTermMatrix(corpus, control = list(weighting = weightTfIdf))

# Inspect the DTM with TF-IDF weighting
inspect(dtm_tfidf)
