# 12/3 CSC 401 Examples

# Load necessary libraries
library(ggplot2)
library(forecast)  # For the stationarity test

# Set seed for reproducibility
set.seed(123)

# Generate a simple non-stationary time series (e.g., with a trend)
n <- 100
time <- 1:n
trend <- 0.5 * time
noise <- rnorm(n, mean = 0, sd = 1)
ts_data <- trend + noise

# Convert to a time series object
ts_data <- ts(ts_data)

# Plot the original time series
ggplot() +
  geom_line(aes(x = time, y = ts_data), color = "blue") +
  ggtitle("Original Time Series with Trend") +
  xlab("Time") +
  ylab("Value")

# Manually calculate the first difference
diff_ts <- numeric(length(ts_data) - 1)

for (i in 2:length(ts_data)) {
  diff_ts[i - 1] <- ts_data[i] - ts_data[i - 1]
}

# Plot the differenced time series
ggplot() +
  geom_line(aes(x = time[-1], y = diff_ts), color = "red") +
  ggtitle("Differenced Time Series") +
  xlab("Time") +
  ylab("Differenced Value")

# Load the urca package for stationarity tests
library(urca)

# Perform the Augmented Dickey-Fuller (ADF) test on the differenced series
adf_test <- ur.df(diff_ts, type = "drift", selectlags = "AIC")

# Print the test results
summary(adf_test)

# Simulate or load a time series (for example, we'll use the AirPassengers dataset)
data("AirPassengers")
time_series <- AirPassengers

# Parameters for the moving average
window_size <- 12 # 12-month (1-year) moving average

# Calculate the moving average
moving_average <- filter(time_series, rep(1/window_size, window_size), sides = 2)

# Plot the original time series
plot(time_series, main="Original Time Series with Moving Average", col="blue", ylab="Number of Passengers", xlab="Time")

# Add the moving average to the plot
lines(moving_average, col="red", lwd=2)

# Add a legend
legend("topleft", legend=c("Original", "Moving Average"), col=c("blue", "red"), lwd=2)

# Load necessary package
if(!require(TTR)) install.packages("TTR", dependencies=TRUE)
library(TTR)

# Simulate or load a time series (for example, we'll use the AirPassengers dataset)
data("AirPassengers")
time_series <- AirPassengers

# Calculate the Exponential Moving Average
ema <- EMA(time_series, n = 12)  # 12-month EMA

# Plot the original time series
plot(time_series, main="Original Time Series with Exponential Moving Average", col="blue", ylab="Number of Passengers", xlab="Time")

# Add the Exponential Moving Average to the plot
lines(ema, col="red", lwd=2)

# Add a legend
legend("topleft", legend=c("Original", "EMA"), col=c("blue", "red"), lwd=2)

# Example time series data (we'll use the AirPassengers dataset)
data("AirPassengers")
time_series <- AirPassengers

# Function to calculate lags
calculate_lags <- function(series, lag = 1) {
  # Create a vector of NA values to shift the series
  lagged_series <- c(rep(NA, lag), series[1:(length(series) - lag)])
  return(lagged_series)
}

# Calculate a lag of 1 for the time series
lag_1 <- calculate_lags(time_series, lag = 1)

# Display the original and lagged series
head(cbind(time_series, lag_1), 10)

# Create a lag plot (lag 1)
plot(lag_1, time_series, main="Lag Plot (lag = 1)", xlab="Lag 1", ylab="Original Series")
abline(lm(time_series ~ lag_1), col="red") # Adding a regression line

# Function to calculate autocorrelation for a given lag
calculate_autocorrelation <- function(series, lag) {
  lagged_series <- calculate_lags(series, lag)
  # Remove NA values
  valid_indices <- !is.na(lagged_series)
  series <- series[valid_indices]
  lagged_series <- lagged_series[valid_indices]
  
  # Calculate the autocorrelation
  n <- length(series)
  mean_series <- mean(series)
  mean_lagged <- mean(lagged_series)
  
  numerator <- sum((series - mean_series) * (lagged_series - mean_lagged))
  denominator <- sqrt(sum((series - mean_series)^2) * sum((lagged_series - mean_lagged)^2))
  
  autocorrelation <- numerator / denominator
  return(autocorrelation)
}

# Calculate autocorrelation for lag 1
autocorrelation_1 <- calculate_autocorrelation(time_series, lag = 1)
autocorrelation_1

# Function to fit a simple AR(1) model
fit_ar_model <- function(series, lag = 1) {
  lagged_series <- calculate_lags(series, lag)
  valid_indices <- !is.na(lagged_series)
  
  # Use linear regression to fit the model
  ar_model <- lm(series[valid_indices] ~ lagged_series[valid_indices])
  return(ar_model)
}

# Fit an AR(1) model
ar_model <- fit_ar_model(time_series, lag = 1)

# Display the summary of the model
summary(ar_model)

# Make predictions based on the model
predictions <- predict(ar_model)

# Plot original vs predicted values
plot(time_series, type = "l", col = "blue", main = "AR(1) Model Predictions vs Original Series")
lines(c(NA, predictions), col = "red")  # Lag causes first prediction to be NA
legend("topleft", legend=c("Original", "Predicted"), col=c("blue", "red"), lty=1)

# Example time series data (we'll use the AirPassengers dataset)
data("AirPassengers")
time_series <- AirPassengers

# Function to calculate lagged series
calculate_lags <- function(series, lag = 1) {
  lagged_series <- c(rep(NA, lag), series[1:(length(series) - lag)])
  return(lagged_series)
}

# Function to calculate autocorrelation for a given lag
calculate_autocorrelation <- function(series, lag) {
  lagged_series <- calculate_lags(series, lag)
  valid_indices <- !is.na(lagged_series)
  series <- series[valid_indices]
  lagged_series <- lagged_series[valid_indices]
  
  n <- length(series)
  mean_series <- mean(series)
  mean_lagged <- mean(lagged_series)
  
  numerator <- sum((series - mean_series) * (lagged_series - mean_lagged))
  denominator <- sqrt(sum((series - mean_series)^2) * sum((lagged_series - mean_lagged)^2))
  
  autocorrelation <- numerator / denominator
  return(autocorrelation)
}

# Calculate ACF for lags 1, 2, and 3
acf_1 <- calculate_autocorrelation(time_series, lag = 1)
acf_2 <- calculate_autocorrelation(time_series, lag = 2)
acf_3 <- calculate_autocorrelation(time_series, lag = 3)

cat("ACF Lag 1:", acf_1, "\n")
cat("ACF Lag 2:", acf_2, "\n")
cat("ACF Lag 3:", acf_3, "\n")

# PACF calculation using Yule-Walker equations
# PACF for lag 1 is just the ACF for lag 1
pacf_1 <- acf_1

# PACF for lag 2
pacf_2 <- (acf_2 - (acf_1 * acf_1)) / (1 - acf_1^2)

# PACF for lag 3
numerator_3 <- acf_3 - (acf_2 * acf_1) - (acf_1 * acf_2)
denominator_3 <- 1 - (acf_1^2) - (acf_2^2)
pacf_3 <- numerator_3 / denominator_3

cat("PACF Lag 1:", pacf_1, "\n")
cat("PACF Lag 2:", pacf_2, "\n")
cat("PACF Lag 3:", pacf_3, "\n")

# Example: Resetting the Graphics Device and Resizing the Plot Panel
dev.off()  # Reset graphics device
windows()  # Opens a new plotting window (use quartz() on macOS, x11() on Linux)

# Load the AirPassengers dataset
data(AirPassengers)
time_series <- AirPassengers

# Step 1: Visualize the Original Series
plot(time_series, main = "Original AirPassengers Time Series")

# Step 2: Estimate Trend Component with Moving Average
window_size <- 12  # 12 months in a year
trend <- stats::filter(time_series, rep(1/window_size, window_size), sides=2)

# Step 3: Detrend the Series to Estimate Seasonal Component
detrended <- time_series - trend

# Calculate Seasonal Indices
seasonal <- tapply(detrended, cycle(time_series), mean, na.rm=TRUE)

# Expand the seasonal component to the length of the time series
seasonal_full <- rep(seasonal, length.out=length(time_series))

# Step 4: Calculate Residual Component
residual <- time_series - trend - seasonal_full

# Step 5: Visualize the Decomposition
par(mfrow=c(4,1))
plot(time_series, main="Original Series", ylab="")
plot(trend, main="Estimated Trend", ylab="")
plot(seasonal_full, main="Estimated Seasonal Component", ylab="")
plot(residual, main="Estimated Residual Component", ylab="")

