# 9/26 CSC 401 Examples

# Load necessary libraries
library(MASS)

# Define the dataset
data <- mtcars
x <- as.matrix(data[, c("disp", "hp", "drat", "wt", "qsec")])
y <- data$mpg

# Standardize the variables
x <- scale(x)
y <- scale(y)

# Lasso penalty (L1)
lasso_penalty <- function(beta, lambda) {
  lambda * sum(abs(beta))
}

# Ridge penalty (L2)
ridge_penalty <- function(beta, lambda) {
  lambda * sum(beta^2)
}

# Elastic Net penalty
elastic_net_penalty <- function(beta, lambda, alpha = 0.5) {
  alpha * lasso_penalty(beta, lambda) + (1 - alpha) * ridge_penalty(beta, lambda)
}

# Custom penalty (for demonstration)
custom_penalty <- function(beta, lambda) {
  lambda * sum(abs(beta)^1.5)  # Custom non-linear penalty
}

# Generalized loss function
loss_function <- function(beta, x, y, penalty_func, lambda, ...) {
  rss <- sum((y - x %*% beta)^2)
  penalty <- penalty_func(beta, lambda, ...)
  return(rss + penalty)
}

gradient_descent <- function(x, y, penalty_func, lambda, learning_rate = 0.01, max_iter = 1000) {
  n <- nrow(x)
  p <- ncol(x)
  beta <- rep(0, p)
  
  for (i in 1:max_iter) {
    # Calculate the gradient
    residuals <- y - x %*% beta
    gradient <- -2 * t(x) %*% residuals / n
    
    # Apply penalty gradient
    penalty_gradient <- grad_penalty(beta, penalty_func, lambda)
    
    # Update beta
    beta <- beta - learning_rate * (gradient + penalty_gradient)
  }
  
  return(beta)
}

grad_lasso <- function(beta, lambda) {
  lambda * sign(beta)
}

grad_ridge <- function(beta, lambda) {
  2 * lambda * beta
}

grad_elastic_net <- function(beta, lambda, alpha = 0.5) {
  alpha * lambda * sign(beta) + (1 - alpha) * 2 * lambda * beta
}

grad_custom <- function(beta, lambda) {
  lambda * 1.5 * abs(beta)^(0.5) * sign(beta)
}

# Then in the optimization function, call the appropriate gradient function based on the penalty type.
gradient_descent <- function(x, y, penalty_func, lambda, learning_rate = 0.01, max_iter = 1000, alpha = 0.5) {
  n <- nrow(x)
  p <- ncol(x)
  beta <- rep(0, p)
  
  for (i in 1:max_iter) {
    residuals <- y - x %*% beta
    gradient <- -2 * t(x) %*% residuals / n
    
    # Determine the penalty gradient
    penalty_gradient <- switch(
      penalty_func,
      "lasso" = grad_lasso(beta, lambda),
      "ridge" = grad_ridge(beta, lambda),
      "elastic_net" = grad_elastic_net(beta, lambda, alpha),
      "custom" = grad_custom(beta, lambda),
      stop("Unknown penalty function")
    )
    
    beta <- beta - learning_rate * (gradient + penalty_gradient)
  }
  
  return(beta)
}

# Example Usage
system.time({
  beta_lasso <- gradient_descent(x, y, penalty_func = "lasso", lambda = 0.1, learning_rate = 0.01)
})

system.time({
  beta_ridge <- gradient_descent(x, y, penalty_func = "ridge", lambda = 0.1, learning_rate = 0.01)
})

system.time({
  beta_elastic_net <- gradient_descent(x, y, penalty_func = "elastic_net", lambda = 0.1, learning_rate = 0.01)
})

system.time({
  beta_custom <- gradient_descent(x, y, penalty_func = "custom", lambda = 0.1, learning_rate = 0.01)
})

# Lasso using glmnet
library(glmnet)
system.time({
  lasso_model <- glmnet(x, y, alpha = 1)
})

# Ridge using glmnet
system.time({
  ridge_model <- glmnet(x, y, alpha = 0)
})

# Elastic Net using glmnet
system.time({
  elastic_net_model <- glmnet(x, y, alpha = 0.5)
})

# Coordinate Descent for Lasso
coordinate_descent <- function(x, y, penalty_func = "lasso", lambda = 0.1, max_iter = 1000, alpha = 0.5) {
  n <- nrow(x)
  p <- ncol(x)
  beta <- rep(0, p)
  X_transpose <- t(x)
  
  for (iter in 1:max_iter) {
    for (j in 1:p) {
      # Predict current residuals excluding the j-th predictor
      residuals <- y - x %*% beta + x[, j] * beta[j]
      
      # Unregularized update for beta[j]
      rho <- X_transpose[j, ] %*% residuals / n
      
      # Update rule depends on the penalty type
      if (penalty_func == "lasso") {
        beta[j] <- soft_thresholding(rho, lambda)
      } else if (penalty_func == "ridge") {
        beta[j] <- rho / (1 + 2 * lambda)
      } else if (penalty_func == "elastic_net") {
        beta[j] <- soft_thresholding(rho, alpha * lambda) / (1 + (1 - alpha) * 2 * lambda)
      } else if (penalty_func == "custom") {
        beta[j] <- custom_update(rho, lambda)  # Define your custom update rule
      } else {
        stop("Unknown penalty function")
      }
    }
  }
  
  return(beta)
}

# Soft Thresholding Operator for Lasso
soft_thresholding <- function(rho, lambda) {
  sign(rho) * max(0, abs(rho) - lambda)
}

# Custom update for custom penalty (example)
custom_update <- function(rho, lambda) {
  # Example custom update (you can adjust based on your penalty)
  return(rho / (1 + lambda * abs(rho)^0.5))
}

# Example usage
system.time({
  beta_lasso_cd <- coordinate_descent(x, y, penalty_func = "lasso", lambda = 0.1)
})

sgd_penalized <- function(x, y, penalty_func = "lasso", lambda = 0.1, learning_rate = 0.01, max_epochs = 1000, alpha = 0.5) {
  n <- nrow(x)
  p <- ncol(x)
  beta <- rep(0, p)
  
  for (epoch in 1:max_epochs) {
    for (i in sample(1:n, n, replace = FALSE)) {
      xi <- x[i, , drop = FALSE]  # Get i-th data point as a row vector
      yi <- y[i]
      
      # Compute the prediction and residual
      prediction <- xi %*% beta
      residual <- yi - prediction
      
      # Gradient of the loss (for a single data point)
      gradient <- -2 * t(xi) %*% residual
      
      # Penalty gradient
      penalty_gradient <- switch(
        penalty_func,
        "lasso" = grad_lasso(beta, lambda),
        "ridge" = grad_ridge(beta, lambda),
        "elastic_net" = grad_elastic_net(beta, lambda, alpha),
        "custom" = grad_custom(beta, lambda),
        stop("Unknown penalty function")
      )
      
      # Update the coefficients
      beta <- beta - learning_rate * (gradient + penalty_gradient)
    }
  }
  
  return(beta)
}

# Reusing the gradient functions from the previous example
grad_lasso <- function(beta, lambda) {
  lambda * sign(beta)
}

grad_ridge <- function(beta, lambda) {
  2 * lambda * beta
}

grad_elastic_net <- function(beta, lambda, alpha = 0.5) {
  alpha * lambda * sign(beta) + (1 - alpha) * 2 * lambda * beta
}

grad_custom <- function(beta, lambda) {
  # Example custom gradient
  lambda * 1.5 * abs(beta)^(0.5) * sign(beta)
}

# Example usage with the mtcars dataset
# Preparing the data
data <- mtcars
y <- data$mpg
x <- as.matrix(data[, c("cyl", "disp", "hp", "wt", "qsec")])

# Standardize the data
x <- scale(x)
y <- scale(y)

# Running the SGD
set.seed(123)  # Set a seed for reproducibility
system.time({
  beta_sgd <- sgd_penalized(x, y, penalty_func = "lasso", lambda = 0.1, learning_rate = 0.01, max_epochs = 500)
})

# Display the coefficients
beta_sgd

library(splines)

# Example dataset
data <- mtcars
y <- data$mpg
x <- data$wt

# Create B-spline basis for 'x'
k <- 4  # Number of knots
degree <- 3  # Degree of spline
bs_basis <- bs(x, df = k, degree = degree)

# Design matrix (including intercept)
X <- cbind(1, bs_basis)

# Penalty matrix for second derivative (assuming B-splines)
penalty_matrix <- diff(diag(k + degree), differences = 2) %*% t(diff(diag(k + degree), differences = 2))

# Regularization parameter
lambda <- 0.1

# Coordinate Descent for Penalized Spline Regression
coordinate_descent_spline <- function(X, y, lambda = 0.1, max_iter = 1000, tol = 1e-6) {
  p <- ncol(X)
  beta <- rep(0, p)
  
  for (iter in 1:max_iter) {
    beta_old <- beta
    
    for (j in 1:p) {
      residuals <- y - X %*% beta + X[, j] * beta[j]
      
      # Ridge penalty for smoothness (second derivative penalty)
      penalty_term <- lambda * penalty_matrix[j, ] %*% beta
      
      # Update rule (assuming squared loss + ridge penalty)
      beta[j] <- sum(residuals * X[, j]) / sum(X[, j]^2 + lambda * penalty_matrix[j, j])
    }
    
    # Check convergence
    if (sqrt(sum((beta - beta_old)^2)) < tol) {
      break
    }
  }
  
  return(beta)
}

# Example usage
beta_spline <- coordinate_descent_spline(X, y, lambda = 0.1)

# Predictions
y_pred <- X %*% beta_spline

# Plotting
plot(x, y, pch = 19, main = "Penalized Spline Fit")
lines(sort(x), y_pred[order(x)], col = "blue", lwd = 2)

##########################################################

# Load necessary library
library(splines)

# Load the dataset
data(mtcars)

# Define the number of knots to test
knots_values <- c(3, 4, 5, 6)

# Define lambda values for regularization (for LASSO/Ridge, which we’ll not implement here directly)
lambda_values <- c(0.01, 0.1, 1)

# Placeholder for results
results <- data.frame(n_knots = integer(), lambda = numeric(), mse = numeric())

# Perform cross-validation to find the optimal number of knots
set.seed(123)
n <- nrow(mtcars)
fold_indices <- sample(rep(1:5, length = n))

for (n_knots in knots_values) {
  num_basis <- n_knots + 2
  for (lambda in lambda_values) {
    fold_mses <- numeric(5)
    
    for (fold in 1:5) {
      train_indices <- which(fold_indices != fold)
      test_indices <- which(fold_indices == fold)
      
      x_train <- mtcars$mpg[train_indices]
      y_train <- mtcars$hp[train_indices]
      x_test <- mtcars$mpg[test_indices]
      y_test <- mtcars$hp[test_indices]
      
      # Create B-spline basis for training data
      bs_basis_train <- bs(x_train, df = num_basis, degree = 3)
      X_train <- cbind(1, bs_basis_train)
      
      # Fit the model using linear regression (not penalized for this example)
      model <- lm(y_train ~ X_train - 1)
      
      # Create B-spline basis for test data
      bs_basis_test <- bs(x_test, df = num_basis, degree = 3)
      X_test <- cbind(1, bs_basis_test)
      
      y_pred <- predict(model, newdata = data.frame(X_test))
      
      # Calculate mean squared error on the test set
      fold_mses[fold] <- mean((y_test - y_pred)^2)
    }
    
    # Average MSE across all folds
    mean_mse <- mean(fold_mses)
    results <- rbind(results, data.frame(n_knots = n_knots, lambda = lambda, mse = mean_mse))
  }
}

# Find the best number of knots and lambda
best_params <- results[which.min(results$mse), ]
print(best_params)

# Plotting the cross-validation results for spline regression
library(ggplot2)

# Assuming 'results' is the data frame with your cross-validation results
ggplot(results, aes(x = n_knots, y = mse, color = factor(lambda))) +
  geom_line() +
  geom_point() +
  labs(title = "Cross-Validation Results for Spline Regression",
       x = "Number of Knots",
       y = "Mean Squared Error (MSE)",
       color = "Lambda") +
  theme_minimal()

# Load necessary library
library(glmnet)

# Prepare the data
x <- as.matrix(mtcars[, c("mpg", "cyl", "disp", "hp", "wt")])
y <- mtcars$mpg

# Perform cross-validation to find the optimal lambda for LASSO
cv_model <- cv.glmnet(x, y, alpha = 1)  # alpha = 1 for LASSO, 0 for Ridge

# Get the best lambda
best_lambda <- cv_model$lambda.min
print(best_lambda)

# Fit the final model with the best lambda
final_model <- glmnet(x, y, alpha = 1, lambda = best_lambda)

# Print model coefficients
print(coef(final_model))

# Load necessary library
library(glmnet)

# Assuming you've run the cross-validation for `glmnet` as shown previously
plot(cv_model)
title("Cross-Validation for Lambda in LASSO (glmnet)")

# Suppose you have the best MSEs from spline regression and glmnet
best_spline_mse <- min(results$mse)
best_glmnet_mse <- min(cv_model$cvm)

# Data frame for comparison
comparison_df <- data.frame(
  Model = c("Spline Regression", "Penalized Regression (LASSO)"),
  MSE = c(best_spline_mse, best_glmnet_mse)
)

# Plotting the comparison
ggplot(comparison_df, aes(x = Model, y = MSE)) +
  geom_bar(stat = "identity", fill = "skyblue") +
  labs(title = "Comparison of Best MSE Between Spline Regression and \nPenalized Regression",
       x = "Model",
       y = "Mean Squared Error (MSE)") +
  theme_minimal()

